/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 03/07/22.
//

#include "topdown.hpp"

#include "fggl/data/storage.hpp"
#include "fggl/gfx/camera.hpp"
#include "fggl/input/camera_input.hpp"
#include "fggl/entity/loader/loader.hpp"

static const fggl::util::GUID collectablePrefab = "collectable"_fid;
static const fggl::util::GUID WallNPrefab = "wallX"_fid;
static const fggl::util::GUID WallEPrefab = "wallZ"_fid;
static const fggl::util::GUID floorPrefab = "floor"_fid;

namespace demo {

	static void create_topdown_camera(fggl::entity::EntityManager& world) {
		auto prototype = world.create();

		// setup camera position/transform
		auto& transform = world.add<fggl::math::Transform>(prototype);
		transform.origin(glm::vec3(10.0f, 50.0f, 10.0f));

		// setup camera components
		auto& camera = world.add<fggl::gfx::Camera>(prototype);
		camera.target = glm::vec3(0.0f, 0.0f, 0.0f);

		auto& cameraKeys = world.add<fggl::input::FreeCamKeys>(prototype);
		cameraKeys.forward = glfwGetKeyScancode(GLFW_KEY_W);
		cameraKeys.backward = glfwGetKeyScancode(GLFW_KEY_S);
		cameraKeys.left = glfwGetKeyScancode(GLFW_KEY_A);
		cameraKeys.right = glfwGetKeyScancode(GLFW_KEY_D);
		cameraKeys.rotate_cw = glfwGetKeyScancode(GLFW_KEY_Q);
		cameraKeys.rotate_ccw = glfwGetKeyScancode(GLFW_KEY_E);
	}

	static void place_cover_boxes(fggl::entity::EntityFactory* factory, fggl::entity::EntityManager& world) {
		std::array<fggl::math::vec3,8> boxPos = {{
													 {-10.0F, 0.0F, -10.0F},
													 {-10.0F, 0.0F,  10.0F},
													 { 10.0F, 0.0F, -10.0F},
													 { 10.0F, 0.0F,  10.0F},
													 {-10.0F, 0.0F,   0.0F},
													 {  0.0F, 0.0F, -10.0F},
													 { 10.0F, 0.0F,   0.0F},
													 {  0.0F, 0.0F,  10.0F},
												 }};
		for (auto pos : boxPos) {
			auto box = factory->create(collectablePrefab, world);
			auto& transform = world.get<fggl::math::Transform>(box);
			transform.origin(pos);
		}
	}

	static void build_arena(fggl::entity::EntityFactory* factory, fggl::entity::EntityManager& world) {
		{
			auto floor = factory->create(floorPrefab, world);
			auto& transform = world.get<fggl::math::Transform>(floor);
			transform.origin({0.0F, -2.5F, 0.0F});
			//fggl::debug::log("created floor: {}", floor);
		}

		fggl::math::vec2 size{40.0F, 40.0F};
		for (auto side : {-1.0F, 1.0F})
		{
			{
				auto obj = factory->create(WallNPrefab, world);
				auto& transform = world.get<fggl::math::Transform>(obj);
				transform.origin({size.x / 2 * side, 0.0F, 0.0F});
			}
			{
				auto obj = factory->create(WallEPrefab, world);
				auto& transform = world.get<fggl::math::Transform>(obj);
				transform.origin({0.0F, 0.0F, size.y/2 * side });
			}
		}

		place_cover_boxes(factory, world);
	}

static void process_camera(fggl::entity::EntityManager& ecs, const fggl::input::Input& input) {
	auto cameras = ecs.find<fggl::gfx::Camera>();
	if ( !cameras.empty() ) {
		auto cam = cameras[0];
		fggl::input::process_scroll(ecs, input, cam);
		fggl::input::process_freecam(ecs, input, cam);
		fggl::input::process_edgescroll(ecs, input, cam);
	}
}

static void populate_sample_level(fggl::entity::EntityFactory* factory, fggl::entity::EntityManager& world) {
	create_topdown_camera(world);
	build_arena(factory, world);
}

TopDown::TopDown(fggl::App& app) : fggl::scenes::Game(app) {

}

void TopDown::activate() {
	Game::activate();

	fggl::debug::log(fggl::debug::Level::info, "TopDown::activate()");
	auto* assetLoader = m_owner.service<fggl::assets::Loader>();
	assetLoader->load("topdown.yml", fggl::entity::PROTOTYPE_ASSET);

	auto* factory = m_owner.service<fggl::entity::EntityFactory>();
	//fggl::ecs3::load_prototype_file(world(), *storage, "topdown.yml");

	// create a sample level
	populate_sample_level(factory, world());
}

void TopDown::update(float dt) {
	Game::update(dt);

	process_camera(world(), input());
	if ( input().mouse.pressed(fggl::input::MouseButton::LEFT) ) {
		pick_object();
	}
}

void TopDown::pick_object() {
		if (!hasPhys()) {
			return;
		}

	auto cameras = world().find<fggl::gfx::Camera>();
	if ( cameras.empty() ) {
		return;
	}

	auto cam = cameras[0];

	fggl::math::vec2 position {
		input().mouse.axis(fggl::input::MouseAxis::X),
		input().mouse.axis(fggl::input::MouseAxis::Y),
	};

	auto ray = fggl::gfx::get_camera_ray(world(), cam, position);
	auto hit = phys().raycast(ray);
	if ( hit != fggl::entity::INVALID) {
		fggl::debug::info("hit: {}", (int)hit);
	} else {
		fggl::debug::info("no hit");
	}
}

void TopDown::render(fggl::gfx::Graphics& gfx) {
	Game::render(gfx);
}

}