/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 03/09/22.
//

#ifndef FGGL_ANIMATION_ANIMATOR_H
#define FGGL_ANIMATION_ANIMATOR_H

#include <functional>
#include <map>
#include <cstdint>
#include <cassert>

namespace fggl::animation {

	using AnimationCallback = std::function<void(void)>;
	using CallbackHandle = uint32_t;

	/**
	 * Frame-based animation.
	 *
	 * Tries to maintain a constant framerate for things that care about that.
	 */
	class FrameAnimator {
		public:
			explicit inline FrameAnimator(float targetFPS) : m_target( 1.0F / targetFPS) {
				assert( 0 <= m_target );
			}

			inline void reset() {
				m_current = 0;
			}

			inline void update(float dt) {
				assert(0 <= dt);
				m_current += dt;
				while ( m_current >= m_target) {
					tick();
					m_current -= m_target;
				}
				assert(0 <= m_current);
			}

			// tick the animation system, should be handled by update
			inline void tick() {
				for (auto& [k,v] : m_callbacks) {
					v();
				}
			}

			inline CallbackHandle add(AnimationCallback callback) {
				auto myHandle = m_lastCallback++;
				m_callbacks[myHandle] = callback;
				return myHandle;
			}

			inline void remove(CallbackHandle handle) {
				auto itr = m_callbacks.find(handle);
				if ( itr != m_callbacks.end() ) {
					m_callbacks.erase(itr);
				}
			}

		private:
			const float m_target;
			float m_current = 0.0F;

			CallbackHandle m_lastCallback = 0;
			std::map<CallbackHandle, AnimationCallback> m_callbacks;
	};

} // namespace fggl::animation

#endif //FGGL_ANIMATION_ANIMATOR_H
