/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Wrappers for types that probably should be implemented, but can be faked for now.
// These are placeholders for the more advanced/specialist data structures we should be using, but mocked out using
// more basic ones.
//

#ifndef FGGL_DS_PLACEHOLDER_HPP
#define FGGL_DS_PLACEHOLDER_HPP

#include <map>
#include <cassert>

namespace fggl::ds {

	template<typename T, std::size_t N>
	class FakeSlotMap {
		public:
			using WeakRef = std::size_t;
			constexpr static WeakRef BAD_INDEX = 0;

			inline bool valid(WeakRef idx) const {
				return m_data.find(idx) != m_data.end();
			}

			WeakRef allocate() {
				if (N <= m_data.size()) {
					assert(0 && "Fake slot map emulated out of space");
					return BAD_INDEX;
				}

				auto myIdx = m_nextIdx++;
				m_data[myIdx] = T();
			}

			void free(WeakRef idx) {
				m_data.erase(idx);
			}

			T &get(WeakRef idx) const {
				assert(valid(idx));
				return m_data[idx];
			}

			T *tryGet(WeakRef idx) const {
				if (valid(idx)) {
					return &m_data[idx];
				}
				return nullptr;
			}

		private:
			std::map<WeakRef, T> m_data;
			std::size_t m_nextIdx = 1;
	};

} // namespace fggl::ds

#endif //FGGL_DS_PLACEHOLDER_HPP
