/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 27/06/22.
//

#ifndef FGGL_MODULES_MANAGER_HPP
#define FGGL_MODULES_MANAGER_HPP

#include "fggl/modules/module.hpp"
#include "fggl/debug/logging.hpp"

#include <queue>
#include <vector>
#include <map>
#include <stack>
#include <set>
#include <iostream>

namespace fggl::modules {

	/**
	 * A class used for representing a Directed Acyclic Graph.
	 *
	 * This class is mostly used for establishing the loading order of classes for module loading.
	 *
	 * @tparam T the type being represented
	 */
	template<typename T>
	class DependencyGraph {
		public:
			DependencyGraph() = default;

			/**
			 * Clear all currently stored dependencies.
			 *
			 * This method will result in the dependency graph being empty, with no known modules.
			 */
			void clear() {
				m_dependencies.clear();
			}

			/**
			 * Add a series of dependencies for an entry.
			 *
			 * If the entry does not already exist, this will create the entry before adding the dependencies to it.
			 * If the entry already exists, this will append the provided dependencies to its existing list.
			 *
			 * @param name the entry having dependencies added
			 * @param dependencies the things it depends on
			 */
			void addAll(const T &name, const std::vector<T> &dependencies) {
				auto existing = m_dependencies.find(name);
				if (existing == m_dependencies.end()) {
					m_dependencies[name] = dependencies;
				} else {
					existing->second.insert(existing->second.end(), dependencies.begin(), dependencies.end());
				}
			}

			/**
			 * Add a single dependency to the graph.
			 *
			 * If the entry does not already exist, this will create the entry before adding the dependencies to it.
			 * If the entry already exists, this will append the provided dependencies to its existing list.
			 *
			 * @param name the entry which depends something else
			 * @param depends the thing it depends on
			 */
			void add(const T &name, const T &depends) {
				m_dependencies[name].push_back(depends);
			}

			bool getOrder(std::stack<T> &stack) {
				std::set<T> visited{};

				for (const auto &module : m_dependencies) {
					if (!visited.contains(module.first)) {
						sortUtil(module.first, visited, stack);
					}
				}

				return true;
			}

			bool getOrderRev(std::queue<T> &stack) {
				std::set<T> visited{};

				for (const auto &module : m_dependencies) {
					if (!visited.contains(module.first)) {
						sortUtilRev(module.first, visited, stack);
					}
				}

				return true;
			}

		private:
			std::map<T, std::vector<T>> m_dependencies;

			void sortUtil(T idx, std::set<T> &visited, std::stack<T> &stack) {
				visited.emplace(idx);

				for (auto dep : m_dependencies.at(idx)) {
					if (!visited.contains(dep))
						sortUtil(dep, visited, stack);
				}

				stack.push(idx);
			}

			void sortUtilRev(T idx, std::set<T> &visited, std::queue<T> &stack) {
				visited.emplace(idx);

				for (auto dep : m_dependencies.at(idx)) {
					if (!visited.contains(dep))
						sortUtilRev(dep, visited, stack);
				}

				stack.push(idx);
			}
	};

	/**
	 * Store and initialise modules present in the engine.
	 *
	 * This class is responsible for keeping track of which modules the library user has requested, and ensuring that
	 * their dependencies are loaded in the correct order. Once the dependency graph has been built and instances
	 * created, it is responsible for providing references to these initialised classes.
	 */
	class Manager {
		public:
			Manager() = default;

			inline void addVirtual(const Config &config) {
				assert(!m_locked);

				m_modules[config.name] = config;
				for (const auto &service : config.provides) {
					m_serviceProviders[service] = config.name;
				}
			}

			template<typename T>
			void use() {
				assert(!m_locked);

				Config config{.name = T::name, .provides = {}, .depends = {}};
				for (auto service : T::provides) {
					config.provides.push_back(service);
				}

				for (auto service : T::depends) {
					config.depends.push_back(service);
				}
				config.factory = T::factory;

				addVirtual(config);
			}

			bool buildGraph() {
				// resolve links between modules
				for (auto &moduleItr : m_modules) {
					if (moduleItr.second.depends.empty()) {
						m_dependencies.addAll(moduleItr.first, {});
						continue;
					}

					for (auto &service : moduleItr.second.depends) {

						auto provider = m_serviceProviders.find(service);
						if (provider == m_serviceProviders.end()) {
							debug::log(debug::Level::warning,
									   "{} depends on {}, but nothing provides it",
									   moduleItr.first,
									   service);
							// nothing can provide the service requested, setup is invalid.
							return false;
						}

						m_dependencies.add(moduleItr.first, provider->second);
					}
				}
				return true;
			}

			template<typename T>
			T *get() const {
				assert(m_locked);
				return m_services.template get<T>();
			}

			void resolve() {
				if (!buildGraph()) {
					return;
				}

				std::queue<ModuleIdentifier> stack;
				m_dependencies.getOrderRev(stack);

				while (!stack.empty()) {
					auto nextToInit = stack.front();
					debug::log(debug::Level::info, "Initializing {}", nextToInit);

					auto &module = m_modules.at(nextToInit);
					if (module.factory != nullptr) {
						for (auto &service : module.provides) {
							bool result = module.factory(service, m_services);
							if (!result) {
								debug::log(debug::Level::warning,
										   "{} could not create service {}",
										   nextToInit,
										   service);
							}
						}
					} else {
						debug::log(debug::Level::warning, "{} has no factory - skipping", nextToInit);
					}
					stack.pop();
				}

				debug::log(debug::Level::info, "engine boot complete");
				m_locked = true;
			}

		private:
			bool m_locked = false;
			Services m_services;
			std::map<ModuleIdentifier, Config> m_modules;
			DependencyGraph<ModuleIdentifier> m_dependencies;
			std::map<ModuleService, ModuleIdentifier> m_serviceProviders;

	};

} // namespace fggl::modules

#endif //FGGL_MODULES_MANAGER_HPP
