#include <gtest/gtest.h>
#include <gmock/gmock.h>

#include <fggl/ecs3/fast/ecs.hpp>

using fggl::ecs::Component;

namespace {

	class DummyComp4 {
		public:
			constexpr static const char* name = "DummyComp2";
			bool a;
			bool b;
			bool c;
	};

	class DummyComp3 {
		public:
			constexpr static const char* name = "DummyComp2";
			int x;
			DummyComp3() = default;
			DummyComp3(int tmp) : x(tmp) {
			}
			DummyComp3( const DummyComp3& other ) : x(other.x) {}
			DummyComp3( DummyComp3&& other ) : x(other.x) {}
	};

	TEST(ecs3, TypeRegistration) {
		fggl::ecs3::TypeRegistry reg;
		reg.make<DummyComp3>();

		EXPECT_EQ(true, reg.exists<DummyComp3>());
		EXPECT_EQ(false, reg.exists<DummyComp4>());
	}

	TEST(ecs3, TypeIDOrdered) {
		auto c1 = fggl::ecs::Component<DummyComp3>::typeID();
		auto c2 = fggl::ecs::Component<DummyComp4>::typeID();

		auto t1 = fggl::ecs3::make_id(2, c1, c2);
		auto t2 = fggl::ecs3::make_id(2, c1, c2);

		EXPECT_EQ( t1, t2 );
		EXPECT_EQ( t2, t1 );
	}

	TEST(ecs3, TypeIDUnordered) {
		auto c1 = fggl::ecs::Component<DummyComp3>::typeID();
		auto c2 = fggl::ecs::Component<DummyComp4>::typeID();

		auto t1 = fggl::ecs3::make_id(2, c1, c2);
		auto t2 = fggl::ecs3::make_id(2, c2, c1);

		EXPECT_EQ( t1, t2 );
		EXPECT_EQ( t2, t1 );
	}

	TEST(ecs3, TypeIDDifferent) {
		auto c1 = fggl::ecs::Component<DummyComp3>::typeID();
		auto c2 = fggl::ecs::Component<DummyComp4>::typeID();

		auto t1 = fggl::ecs3::make_id(2, c1, c2);
		auto t2 = fggl::ecs3::make_id(1, c2);

		EXPECT_NE( t1, t2 );
		EXPECT_NE( t2, t1 );
	}

	TEST(ecs3, createSingle) {
		fggl::ecs3::TypeRegistry reg;
		reg.make<DummyComp3>();

		auto type = fggl::ecs3::make_id(1, fggl::ecs::Component<DummyComp3>::typeID());
		fggl::ecs3::Container container(reg, type);

		auto pos = container.create();
        auto* data = container.data<DummyComp3>();
        data[pos].x = 42;

        ASSERT_EQ(1, container.size());
	}

    TEST(ecs3, createTwo) {
        fggl::ecs3::TypeRegistry reg;
        reg.make<DummyComp3>();
        reg.make<DummyComp4>();

        auto type = fggl::ecs3::make_id(2, fggl::ecs::Component<DummyComp3>::typeID(), fggl::ecs::Component<DummyComp4>::typeID());
        fggl::ecs3::Container container(reg, type);

        // create object
        auto pos = container.create();
        auto* data = container.data<DummyComp3>();
        data[pos].x = 42;

        ASSERT_EQ(1, container.size());
    }

	TEST(ecs3, createRemoveSingle) {
		fggl::ecs3::TypeRegistry reg;
		reg.make<DummyComp3>();

		auto type = fggl::ecs3::make_id(1, fggl::ecs::Component<DummyComp3>::typeID());
		fggl::ecs3::Container container(reg, type);

        // create object
        auto pos = container.create();
        auto* data = container.data<DummyComp3>();
        data[pos].x = 42;
        container.remove(pos);

        ASSERT_EQ(0, container.size());
	}

    TEST(ecs3, createRemoveTwo) {
        fggl::ecs3::TypeRegistry reg;
        reg.make<DummyComp3>();
        reg.make<DummyComp4>();

        auto type = fggl::ecs3::make_id(2, fggl::ecs::Component<DummyComp3>::typeID(), fggl::ecs::Component<DummyComp4>::typeID());
        fggl::ecs3::Container container(reg, type);

        auto pos = container.create();
        auto* data = container.data<DummyComp3>();
        data[pos].x = 42;
        container.remove(pos);

        ASSERT_EQ(0, container.size());
    }

	TEST(ecs3, createRemoveLast) {
		fggl::ecs3::TypeRegistry reg;
		reg.make<DummyComp3>();
		reg.make<DummyComp4>();

		auto type = fggl::ecs3::make_id(1, fggl::ecs::Component<DummyComp3>::typeID());
		fggl::ecs3::Container container(reg, type);

        auto pos = container.create();
        auto pos2 = container.create();

        auto* data = container.data<DummyComp3>();
        data[pos].x = 0xFEED;
        data[pos2].x = 0xBEEF;

		container.remove(pos2);

        // remove may invalidate data pointer
        data = container.data<DummyComp3>();
        ASSERT_EQ(data[pos].x, 0XFEED);
        ASSERT_EQ(1, container.size());
	}

    TEST(ecs3, createRemoveFirst) {
		fggl::ecs3::TypeRegistry reg;
		reg.make<DummyComp3>();

		auto type = fggl::ecs3::make_id(1, fggl::ecs::Component<DummyComp3>::typeID());
		fggl::ecs3::Container container(reg, type);

		auto pos = container.create();
		auto pos2 = container.create();

        auto* data = container.data<DummyComp3>();
        data[pos].x = 0xFEED;
        data[pos2].x = 0xBEEF;

        container.remove(pos);

        // remove may invalidate data pointer
        data = container.data<DummyComp3>();
        ASSERT_EQ(data[pos].x, 0XBEEF );
        ASSERT_EQ(1, container.size());
	}

    TEST(ecs3, expandFirst) {
        fggl::ecs3::TypeRegistry reg;
        reg.make<DummyComp3>();
        reg.make<DummyComp4>();

        auto type1 = fggl::ecs3::make_id(1, fggl::ecs::Component<DummyComp3>::typeID());
        auto type2 = fggl::ecs3::make_id(2, fggl::ecs::Component<DummyComp3>::typeID(), fggl::ecs::Component<DummyComp4>::typeID());

        fggl::ecs3::Container container1(reg, type1);
        fggl::ecs3::Container container2(reg, type2);

        auto p1 = container1.create();
        auto p2 = container1.create();
        auto data1 = container1.data<DummyComp3>();

        data1[p1].x = 0xFEED;
        data1[p2].x = 0xABCD;

        container2.expand(container1, p1, fggl::ecs::Component<DummyComp4>::typeID());
        data1 = container1.data<DummyComp3>();
        auto data2 = container2.data<DummyComp3>();

        ASSERT_EQ(1, container1.size() );
        ASSERT_EQ(1, container2.size() );

        ASSERT_EQ(0xABCD, data1[0].x);
        ASSERT_EQ(0xFEED, data2[0].x);
    }

    TEST(ecs3, expandLast) {
        fggl::ecs3::TypeRegistry reg;
        reg.make<DummyComp3>();
        reg.make<DummyComp4>();

        auto type1 = fggl::ecs3::make_id(1, fggl::ecs::Component<DummyComp3>::typeID());
        auto type2 = fggl::ecs3::make_id(2, fggl::ecs::Component<DummyComp3>::typeID(), fggl::ecs::Component<DummyComp4>::typeID());

        fggl::ecs3::Container container1(reg, type1);
        fggl::ecs3::Container container2(reg, type2);

        auto p1 = container1.create();
        auto p2 = container1.create();
        auto data1 = container1.data<DummyComp3>();

        data1[p1].x = 0xFEED;
        data1[p2].x = 0xABCD;

        container2.expand(container1, p2, fggl::ecs::Component<DummyComp4>::typeID());
        data1 = container1.data<DummyComp3>();
        auto data2 = container2.data<DummyComp3>();

        ASSERT_EQ(1, container1.size() );
        ASSERT_EQ(1, container2.size() );

        ASSERT_EQ(0xABCD, data2[0].x);
        ASSERT_EQ(0xFEED, data1[0].x);
    }
}
